from bbot.modules.shodan_dns import shodan_dns


class passivetotal(shodan_dns):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {"description": "Query the PassiveTotal API for subdomains", "auth_required": True}
    options = {"username": "", "api_key": ""}
    options_desc = {"username": "RiskIQ Username", "api_key": "RiskIQ API Key"}

    base_url = "https://api.passivetotal.org/v2"

    async def setup(self):
        self.username = self.config.get("username", "")
        self.api_key = self.config.get("api_key", "")
        self.auth = (self.username, self.api_key)
        return await super().setup()

    async def ping(self):
        url = f"{self.base_url}/account/quota"
        j = (await self.request_with_fail_count(url, auth=self.auth)).json()
        limit = j["user"]["limits"]["search_api"]
        used = j["user"]["counts"]["search_api"]
        assert used < limit, "No quota remaining"

    def abort_if(self, event):
        # RiskIQ is famous for their junk data
        return super().abort_if(event) or "unresolved" in event.tags

    async def request_url(self, query):
        url = f"{self.base_url}/enrichment/subdomains?query={self.helpers.quote(query)}"
        return await self.request_with_fail_count(url, auth=self.auth)

    def parse_results(self, r, query):
        for subdomain in r.json().get("subdomains", []):
            yield f"{subdomain}.{query}"

    @property
    def auth_secret(self):
        return self.username and self.api_key



from .base import ModuleTestBase


class TestZoomEye(ModuleTestBase):
    config_overrides = {"modules": {"zoomeye": {"api_key": "asdf", "include_related": True, "max_pages": 3}}}

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://api.zoomeye.org/resources-info",
            match_headers={"API-KEY": "asdf"},
            json={"quota_info": {"remain_total_quota": 5}},
        )
        module_test.httpx_mock.add_response(
            url="https://api.zoomeye.org/domain/search?q=blacklanternsecurity.com&type=0&page=1",
            json={"list": [{"name": "asdf.blacklanternsecurity.com"}]},
        )
        module_test.httpx_mock.add_response(
            url="https://api.zoomeye.org/domain/search?q=blacklanternsecurity.com&type=0&page=2",
            json={"list": [{"name": "zzzz.blacklanternsecurity.com"}]},
        )
        module_test.httpx_mock.add_response(
            url="https://api.zoomeye.org/domain/search?q=blacklanternsecurity.com&type=0&page=3",
            json={"list": [{"name": "ffff.blacklanternsecurity.com"}, {"name": "affiliate.bls"}]},
        )
        module_test.httpx_mock.add_response(
            url="https://api.zoomeye.org/domain/search?q=blacklanternsecurity.com&type=0&page=4",
            json={"list": [{"name": "nope.blacklanternsecurity.com"}]},
        )

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect subdomain #1"
        assert any(e.data == "zzzz.blacklanternsecurity.com" for e in events), "Failed to detect subdomain #2"
        assert any(e.data == "ffff.blacklanternsecurity.com" for e in events), "Failed to detect subdomain #3"
        assert any(e.data == "affiliate.bls" and "affiliate" in e.tags for e in events), "Failed to detect affiliate"
        assert not any(e.data == "nope.blacklanternsecurity.com" for e in events), "Failed to obey max_pages"



from .test_module_paramminer_headers import *


class TestParamminer_Getparams(TestParamminer_Headers):
    modules_overrides = ["httpx", "paramminer_getparams"]
    config_overrides = {"modules": {"paramminer_getparams": {"wordlist": tempwordlist(["canary", "id"])}}}
    getparam_body = """
    <html>
    <title>the title</title>
    <body>
    <p>Hello null!</p>';
    </body>
    </html>
    """

    getparam_body_match = """
    <html>
    <title>the title</title>
    <body>
    <p>Hello AAAAAAAAAAAAAA!</p>';
    </body>
    </html>
    """

    async def setup_after_prep(self, module_test):
        module_test.scan.modules["paramminer_getparams"].rand_string = lambda *args, **kwargs: "AAAAAAAAAAAAAA"
        module_test.monkeypatch.setattr(
            helper.HttpCompare, "gen_cache_buster", lambda *args, **kwargs: {"AAAAAA": "1"}
        )
        expect_args = {"query_string": b"id=AAAAAAAAAAAAAA&AAAAAA=1"}
        respond_args = {"response_data": self.getparam_body_match}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        respond_args = {"response_data": self.getparam_body}
        module_test.set_expect_requests(respond_args=respond_args)

    def check(self, module_test, events):
        assert any(
            e.type == "FINDING" and e.data["description"] == "[Paramminer] Getparam: [id] Reasons: [body]"
            for e in events
        )
        assert not any(
            e.type == "FINDING" and e.data["description"] == "[Paramminer] Getparam: [canary] Reasons: [body]"
            for e in events
        )



from .paramminer_headers import paramminer_headers


class paramminer_cookies(paramminer_headers):
    """
    Inspired by https://github.com/PortSwigger/param-miner
    """

    watched_events = ["URL"]
    produced_events = ["FINDING"]
    flags = ["active", "aggressive", "slow", "web-paramminer"]
    meta = {
        "description": "Smart brute-force to check for common HTTP cookie parameters",
    }
    options = {"wordlist": "https://raw.githubusercontent.com/PortSwigger/param-miner/master/resources/params"}
    options_desc = {"wordlist": "Define the wordlist to be used to derive cookies"}
    scanned_hosts = []
    cookie_blacklist = []
    max_event_handlers = 12
    in_scope_only = True
    compare_mode = "cookie"

    async def check_batch(self, compare_helper, url, cookie_list):
        cookies = {p: self.rand_string(14) for p in cookie_list}
        return await compare_helper.compare(url, cookies=cookies)

    def gen_count_args(self, url):
        cookie_count = 40
        while 1:
            if cookie_count < 0:
                break
            fake_cookies = {self.rand_string(14): self.rand_string(14) for _ in range(0, cookie_count)}
            yield cookie_count, (url,), {"cookies": fake_cookies}
            cookie_count -= 5

    def clean_list(self, cookie):
        if (len(cookie) > 0) and (cookie.strip() not in self.cookie_blacklist):
            return True
        return False



from bbot.modules.shodan_dns import shodan_dns


class censys(shodan_dns):
    """
    thanks to https://github.com/owasp-amass/amass/blob/master/resources/scripts/cert/censys.ads
    """

    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME", "EMAIL_ADDRESS", "IP_ADDRESS", "OPEN_PORT", "PROTOCOL"]
    flags = ["subdomain-enum", "email-enum", "passive", "safe"]
    meta = {"description": "Query the Censys API", "auth_required": True}
    options = {"api_id": "", "api_secret": ""}
    options_desc = {"api_id": "Censys.io API ID", "api_secret": "Censys.io API Secret"}

    base_url = "https://search.censys.io/api/v1"

    async def setup(self):
        self.api_id = self.config.get("api_id", "")
        self.api_secret = self.config.get("api_secret", "")
        self.auth = (self.api_id, self.api_secret)
        return await super().setup()

    async def ping(self):
        url = f"{self.base_url}/account"
        resp = await self.helpers.request(url, auth=self.auth)
        d = resp.json()
        assert isinstance(d, dict), f"Invalid response from {url}: {resp}"
        quota = d.get("quota", {})
        used = int(quota.get("used", 0))
        allowance = int(quota.get("allowance", 0))
        assert used < allowance, "No quota remaining"

    async def query(self, query):
        results = set()
        page = 1
        while 1:
            resp = await self.helpers.request(
                f"{self.base_url}/search/certificates",
                method="POST",
                json={
                    "query": f"parsed.names: {query}",
                    "page": page,
                    "fields": ["parsed.names"],
                },
                auth=self.auth,
            )
            page += 1

            if resp is None:
                break

            d = resp.json()
            if d is None:
                break
            elif not isinstance(d, dict):
                break

            error = d.get("error", "")
            if error:
                self.warning(error)

            if resp.status_code < 200 or resp.status_code >= 400:
                break

            elif d.get("status") is None or d["status"] != "ok" or len(d.get("results", [])) == 0:
                break

            for r in d["results"]:
                for v in r["parsed.names"]:
                    results.add(v.strip(".*").lower())

            metadata = d.get("metadata", {})
            if metadata.get("page", 0) >= metadata.get("pages", 0):
                break

        return results

    @property
    def auth_secret(self):
        return self.api_id and self.api_secret



import re

from ..bbot_fixtures import *


@pytest.mark.asyncio
async def test_web_helpers(bbot_scanner, bbot_config, bbot_httpserver):
    scan1 = bbot_scanner("8.8.8.8", config=bbot_config)
    scan2 = bbot_scanner("127.0.0.1", config=bbot_config)

    user_agent = bbot_config.get("user_agent", "")
    headers = {"User-Agent": user_agent}
    custom_headers = bbot_config.get("http_headers", {})
    headers.update(custom_headers)
    assert headers["test"] == "header"

    url = bbot_httpserver.url_for("/test_http_helpers")
    # test user agent + custom headers
    bbot_httpserver.expect_request(uri="/test_http_helpers", headers=headers).respond_with_data(
        "test_http_helpers_yep"
    )
    response = await scan1.helpers.request(url)
    # should fail because URL is not in-scope
    assert response.status_code == 500
    response = await scan2.helpers.request(url)
    # should suceed because URL is in-scope
    assert response.status_code == 200
    assert response.text == "test_http_helpers_yep"

    # download file
    path = "/test_http_helpers_download"
    url = bbot_httpserver.url_for(path)
    download_content = "test_http_helpers_download_yep"
    bbot_httpserver.expect_request(uri=path).respond_with_data(download_content)
    filename = await scan1.helpers.download(url)
    assert Path(str(filename)).is_file()
    assert scan1.helpers.is_cached(url)
    with open(filename) as f:
        assert f.read() == download_content
    # 404
    path = "/test_http_helpers_download_404"
    url = bbot_httpserver.url_for(path)
    download_content = "404"
    bbot_httpserver.expect_request(uri=path).respond_with_data(download_content, status=404)
    filename = await scan1.helpers.download(url)
    assert filename is None
    assert not scan1.helpers.is_cached(url)

    # wordlist
    path = "/test_http_helpers_wordlist"
    url = bbot_httpserver.url_for(path)
    download_content = "a\ncool\nword\nlist"
    bbot_httpserver.expect_request(uri=path).respond_with_data(download_content)
    filename = await scan1.helpers.wordlist(url)
    assert Path(str(filename)).is_file()
    assert scan1.helpers.is_cached(url)
    assert list(scan1.helpers.read_file(filename)) == ["a", "cool", "word", "list"]

    # page iteration
    base_path = "/test_http_page_iteration"
    template_path = base_path + "/{page}?page_size={page_size}&offset={offset}"
    template_url = bbot_httpserver.url_for(template_path)
    bbot_httpserver.expect_request(
        uri=f"{base_path}/1", query_string={"page_size": "100", "offset": "0"}
    ).respond_with_data("page1")
    bbot_httpserver.expect_request(
        uri=f"{base_path}/2", query_string={"page_size": "100", "offset": "100"}
    ).respond_with_data("page2")
    bbot_httpserver.expect_request(
        uri=f"{base_path}/3", query_string={"page_size": "100", "offset": "200"}
    ).respond_with_data("page3")
    results = []
    agen = scan1.helpers.api_page_iter(template_url)
    try:
        async for result in agen:
            if result and result.text.startswith("page"):
                results.append(result)
            else:
                break
    finally:
        await agen.aclose()
    assert not results
    agen = scan1.helpers.api_page_iter(template_url, json=False)
    try:
        async for result in agen:
            if result and result.text.startswith("page"):
                results.append(result)
            else:
                break
    finally:
        await agen.aclose()
    assert [r.text for r in results] == ["page1", "page2", "page3"]


@pytest.mark.asyncio
async def test_web_interactsh(bbot_scanner, bbot_config, bbot_httpserver):
    from bbot.core.helpers.interactsh import server_list

    scan1 = bbot_scanner("8.8.8.8", config=bbot_config)

    interactsh_client = scan1.helpers.interactsh()

    async def async_callback(data):
        log.debug(f"interactsh poll: {data}")

    interactsh_domain = await interactsh_client.register(callback=async_callback)
    url = f"https://{interactsh_domain}/bbot_interactsh_test"
    response = await scan1.helpers.request(url)
    assert response.status_code == 200
    await asyncio.sleep(10)
    assert any(interactsh_domain.endswith(f"{s}") for s in server_list)
    data_list = await interactsh_client.poll()
    assert isinstance(data_list, list)
    assert any("bbot_interactsh_test" in d.get("raw-request", "") for d in data_list)
    assert await interactsh_client.deregister() is None


@pytest.mark.asyncio
async def test_web_curl(bbot_scanner, bbot_config, bbot_httpserver):
    scan = bbot_scanner("127.0.0.1", config=bbot_config)
    helpers = scan.helpers
    url = bbot_httpserver.url_for("/curl")
    bbot_httpserver.expect_request(uri="/curl").respond_with_data("curl_yep")
    bbot_httpserver.expect_request(uri="/index.html").respond_with_data("curl_yep_index")
    assert await helpers.curl(url=url) == "curl_yep"
    assert await helpers.curl(url=url, ignore_bbot_global_settings=True) == "curl_yep"
    assert (await helpers.curl(url=url, head_mode=True)).startswith("HTTP/")
    assert await helpers.curl(url=url, raw_body="body") == "curl_yep"
    assert (
        await helpers.curl(
            url=url,
            raw_path=True,
            headers={"test": "test", "test2": ["test2"]},
            ignore_bbot_global_settings=False,
            post_data={"test": "test"},
            method="POST",
            cookies={"test": "test"},
            path_override="/index.html",
        )
        == "curl_yep_index"
    )
    # test custom headers
    bbot_httpserver.expect_request("/test-custom-http-headers-curl", headers={"test": "header"}).respond_with_data(
        "curl_yep_headers"
    )
    headers_url = bbot_httpserver.url_for("/test-custom-http-headers-curl")
    curl_result = await helpers.curl(url=headers_url)
    assert curl_result == "curl_yep_headers"


@pytest.mark.asyncio
async def test_web_http_compare(httpx_mock, helpers):
    httpx_mock.add_response(url=re.compile(r"http://www\.example\.com.*"), text="wat")
    compare_helper = helpers.http_compare("http://www.example.com")
    await compare_helper.compare("http://www.example.com", headers={"asdf": "asdf"})
    await compare_helper.compare("http://www.example.com", cookies={"asdf": "asdf"})
    await compare_helper.compare("http://www.example.com", check_reflection=True)
    compare_helper.compare_body({"asdf": "fdsa"}, {"fdsa": "asdf"})
    for mode in ("getparam", "header", "cookie"):
        assert await compare_helper.canary_check("http://www.example.com", mode=mode) == True



from .base import ModuleTestBase


class TestNmap(ModuleTestBase):
    targets = ["127.0.0.1"]
    config_overrides = {"modules": {"nmap": {"ports": "8888,8889"}}}

    def check(self, module_test, events):
        assert any(e.data == "127.0.0.1:8888" for e in events)
        assert not any(e.data == "127.0.0.1:8889" for e in events)



from bbot.modules.bucket_aws import bucket_aws


class bucket_firebase(bucket_aws):
    watched_events = ["DNS_NAME", "STORAGE_BUCKET"]
    produced_events = ["STORAGE_BUCKET", "FINDING"]
    flags = ["active", "safe", "cloud-enum", "web-basic", "web-thorough"]
    meta = {"description": "Check for open Firebase databases related to target"}
    options = {"permutations": False}
    options_desc = {
        "permutations": "Whether to try permutations",
    }

    cloud_helper_name = "firebase"
    delimiters = ("", "-")
    base_domains = ["firebaseio.com"]

    async def check_bucket_exists(self, bucket_name, url):
        url = url.strip("/") + "/.json"
        return await super().check_bucket_exists(bucket_name, url)

    async def check_bucket_open(self, bucket_name, url):
        url = url.strip("/") + "/.json"
        response = await self.helpers.request(url)
        tags = self.gen_tags_exists(response)
        status_code = getattr(response, "status_code", 404)
        msg = ""
        if status_code == 200:
            msg = "Open storage bucket"
        return (msg, tags)



from .base import ModuleTestBase


class TestCensys(ModuleTestBase):
    config_overrides = {"modules": {"censys": {"api_id": "api_id", "api_secret": "api_secret"}}}

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://search.censys.io/api/v1/account",
            json={
                "email": "info@blacklanternsecurity.com",
                "login": "nope",
                "first_login": "1917-08-03 20:03:55",
                "last_login": "1918-05-19 01:15:22",
                "quota": {"used": 26, "allowance": 250, "resets_at": "1919-06-03 16:30:32"},
            },
        )
        module_test.httpx_mock.add_response(
            url="https://search.censys.io/api/v1/search/certificates",
            match_content=b'{"query": "parsed.names: blacklanternsecurity.com", "page": 1, "fields": ["parsed.names"]}',
            json={
                "status": "ok",
                "metadata": {
                    "query": "parsed.names: blacklanternsecurity.com",
                    "count": 1,
                    "backend_time": 4465,
                    "page": 1,
                    "pages": 4,
                },
                "results": [
                    {
                        "parsed.names": [
                            "asdf.blacklanternsecurity.com",
                            "zzzz.blacklanternsecurity.com",
                        ]
                    },
                ],
            },
        )

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"
        assert any(e.data == "zzzz.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"



from .base import ModuleTestBase


class TestAsset_Inventory(ModuleTestBase):
    targets = ["8.8.8.8"]
    scan_name = "asset_inventory_test"
    config_overrides = {"dns_resolution": True}
    modules_overrides = ["asset_inventory", "speculate", "sslcert"]

    def check(self, module_test, events):
        assert any(e.type == "OPEN_TCP_PORT" for e in events), "No open port found"
        assert any(e.type == "DNS_NAME" for e in events), "No DNS name found"
        filename = next(module_test.scan.home.glob("asset-inventory.csv"))
        with open(filename) as f:
            assert "8.8.8.8,,8.8.8.8" in f.read()
        filename = next(module_test.scan.home.glob("asset-inventory-ip-addresses-table*.txt"))
        with open(filename) as f:
            assert "8.8.0.0/16" in f.read()
        filename = next(module_test.scan.home.glob("asset-inventory-domains-table*.txt"))
        with open(filename) as f:
            assert "dns.google" in f.read()


class TestAsset_InventoryEmitPrevious(TestAsset_Inventory):
    config_overrides = {"dns_resolution": True, "output_modules": {"asset_inventory": {"use_previous": True}}}
    modules_overrides = ["asset_inventory"]

    def check(self, module_test, events):
        assert any(e.type == "OPEN_TCP_PORT" for e in events), "No open port found"
        assert any(e.type == "DNS_NAME" for e in events), "No DNS name found"
        filename = next(module_test.scan.home.glob("asset-inventory.csv"))
        with open(filename) as f:
            assert "8.8.8.8,,8.8.8.8" in f.read()
        filename = next(module_test.scan.home.glob("asset-inventory-ip-addresses-table*.txt"))
        with open(filename) as f:
            assert "8.8.0.0/16" in f.read()
        filename = next(module_test.scan.home.glob("asset-inventory-domains-table*.txt"))
        with open(filename) as f:
            assert "dns.google" in f.read()



import csv
import ipaddress
from contextlib import suppress

from .csv import CSV
from bbot.core.helpers.misc import make_ip_type, is_ip, is_port

severity_map = {
    "INFO": 0,
    0: "N/A",
    1: "LOW",
    2: "MEDIUM",
    3: "HIGH",
    4: "CRITICAL",
    "N/A": 0,
    "LOW": 1,
    "MEDIUM": 2,
    "HIGH": 3,
    "CRITICAL": 4,
}


class asset_inventory(CSV):
    watched_events = ["OPEN_TCP_PORT", "DNS_NAME", "URL", "FINDING", "VULNERABILITY", "TECHNOLOGY", "IP_ADDRESS"]
    produced_events = ["IP_ADDRESS", "OPEN_TCP_PORT"]
    meta = {"description": "Output to an asset inventory style flattened CSV file"}
    options = {"output_file": "", "use_previous": False, "summary_netmask": 16}
    options_desc = {
        "output_file": "Set a custom output file",
        "use_previous": "Emit previous asset inventory as new events (use in conjunction with -n <old_scan_name>)",
        "summary_netmask": "Subnet mask to use when summarizing IP addresses at end of scan",
    }

    header_row = ["Host", "Provider", "IP(s)", "Status", "Open Ports", "Risk Rating", "Findings", "Description"]
    filename = "asset-inventory.csv"

    async def setup(self):
        self.assets = {}
        self.open_port_producers = "httpx" in self.scan.modules or any(
            ["portscan" in m.flags for m in self.scan.modules.values()]
        )
        self.use_previous = self.config.get("use_previous", False)
        self.summary_netmask = self.config.get("summary_netmask", 16)
        self.emitted_contents = False
        self._ran_hooks = False
        ret = await super().setup()
        return ret

    async def filter_event(self, event):
        if event._internal:
            return False, "event is internal"
        if event.type not in self.watched_events:
            return False, "event type is not in watched_events"
        if not self.scan.in_scope(event):
            return False, "event is not in scope"
        if "unresolved" in event.tags:
            return False, "event is unresolved"
        return True, ""

    async def handle_event(self, event):
        if (await self.filter_event(event))[0]:
            hostkey = _make_hostkey(event.host, event.resolved_hosts)
            if hostkey not in self.assets:
                self.assets[hostkey] = Asset(event.host)
            self.assets[hostkey].absorb_event(event)

    async def report(self):
        stats = dict()
        totals = dict()

        def increment_stat(stat, value):
            try:
                totals[stat] += 1
            except KeyError:
                totals[stat] = 1
            if not stat in stats:
                stats[stat] = {}
            try:
                stats[stat][value] += 1
            except KeyError:
                stats[stat][value] = 1

        def sort_key(asset):
            host = str(asset.host)
            is_digit = False
            with suppress(IndexError):
                is_digit = host[0].isdigit()
            return (is_digit, host)

        for asset in sorted(self.assets.values(), key=sort_key):
            findings_and_vulns = asset.findings.union(asset.vulnerabilities)
            ports = getattr(asset, "ports", set())
            ports = [str(p) for p in sorted([int(p) for p in asset.ports])]
            ips = sorted([str(i) for i in getattr(asset, "ip_addresses", [])])
            host = getattr(asset, "host", "")
            if host:
                domain = self.helpers.tldextract(host).registered_domain
                if domain:
                    increment_stat("Domains", domain)
            for ip in ips:
                net = ipaddress.ip_network(f"{ip}/{self.summary_netmask}", strict=False)
                increment_stat("IP Addresses", str(net))
            for port in ports:
                increment_stat("Open Ports", port)
            row = {
                "Host": host,
                "Provider": getattr(asset, "provider", ""),
                "IP(s)": ",".join(ips),
                "Status": "Active" if asset.ports else "N/A",
                "Open Ports": ",".join(ports),
                "Risk Rating": severity_map[getattr(asset, "risk_rating", "")],
                "Findings": "\n".join(findings_and_vulns),
                "Description": "\n".join(str(x) for x in getattr(asset, "technologies", set())),
            }
            row.update(asset.custom_fields)
            self.writerow(row)

        for header in ("Domains", "IP Addresses", "Open Ports"):
            table_header = [header, ""]
            if header in stats:
                table = []
                stats_sorted = sorted(stats[header].items(), key=lambda x: x[-1], reverse=True)
                total = totals[header]
                for k, v in stats_sorted:
                    table.append([str(k), f"{v:,}/{total} ({v/total*100:.1f}%)"])
                self.log_table(table, table_header, table_name=f"asset-inventory-{header}")

        if self._file is not None:
            self.info(f"Saved asset-inventory output to {self.output_file}")

    async def finish(self):
        if self.use_previous and not self.emitted_contents:
            self.emitted_contents = True
            if self.output_file.is_file():
                self.info(f"Emitting previous results from {self.output_file}")
                with open(self.output_file, newline="") as f:
                    c = csv.DictReader(f)
                    for row in c:
                        host = row.get("Host", "").strip()
                        ips = row.get("IP(s)", "")
                        if not host or not ips:
                            continue
                        hostkey = _make_hostkey(host, ips)
                        asset = self.assets.get(hostkey, None)
                        if asset is None:
                            asset = Asset(host)
                            self.assets[hostkey] = asset
                        asset.absorb_csv_row(row)
                        self.add_custom_headers(list(asset.custom_fields))
                        if not is_ip(asset.host):
                            host_event = self.make_event(asset.host, "DNS_NAME", source=self.scan.root_event)
                            self.emit_event(host_event)
                            for port in asset.ports:
                                netloc = self.helpers.make_netloc(asset.host, port)
                                open_port_event = self.make_event(netloc, "OPEN_TCP_PORT", source=host_event)
                                self.emit_event(open_port_event)
                        else:
                            for ip in asset.ip_addresses:
                                ip_event = self.make_event(ip, "IP_ADDRESS", source=self.scan.root_event)
                                self.emit_event(ip_event)
                                for port in asset.ports:
                                    netloc = self.helpers.make_netloc(ip, port)
                                    open_port_event = self.make_event(netloc, "OPEN_TCP_PORT", source=ip_event)
                                    self.emit_event(open_port_event)
            else:
                self.warning(
                    f"use_previous=True was set but no previous asset inventory was found at {self.output_file}"
                )
        else:
            self._run_hooks()

    def _run_hooks(self):
        """
        modules can use self.asset_inventory_hook() to add custom functionality to asset_inventory
        the asset inventory module is passed in as the first argument to the method.
        """
        if not self._ran_hooks:
            self._ran_hooks = True
            for module in self.scan.modules.values():
                hook = getattr(module, "asset_inventory_hook", None)
                if hook is not None and callable(hook):
                    hook(self)


class Asset:
    def __init__(self, host):
        self.host = host
        self.ip_addresses = set()
        self.ports = set()
        self.findings = set()
        self.vulnerabilities = set()
        self.status = "UNKNOWN"
        self.risk_rating = 0
        self.provider = ""
        self.technologies = set()
        self.custom_fields = {}

    def absorb_csv_row(self, row):
        # host
        host = row.get("Host", "").strip()
        if host and not is_ip(host):
            self.host = host
        # ips
        self.ip_addresses = set(_make_ip_list(row.get("IP(s)", "")))
        # ports
        ports = [i.strip() for i in row.get("Open Ports", "").split(",")]
        self.ports.update(set(i for i in ports if i and is_port(i)))
        # findings
        findings = [i.strip() for i in row.get("Findings", "").splitlines()]
        self.findings.update(set(i for i in findings if i))
        # technologies
        technologies = [i.strip() for i in row.get("Description", "").splitlines()]
        self.technologies.update(set(i for i in technologies if i))
        # risk rating
        risk_rating = row.get("Risk Rating", "").strip()
        if risk_rating and risk_rating.isdigit() and int(risk_rating) > self.risk_rating:
            self.risk_rating = int(risk_rating)
        # provider
        provider = row.get("Provider", "").strip()
        if provider:
            self.provider = provider
        # custom fields
        for k, v in row.items():
            v = str(v)
            # update the custom field if it doesn't clash with our main fields
            # and if the new value isn't blank
            if v and k not in asset_inventory.header_row:
                self.custom_fields[k] = v

    def absorb_event(self, event):
        if not is_ip(event.host):
            self.host = event.host

        self.ip_addresses = set(_make_ip_list(event.resolved_hosts))

        if event.port:
            self.ports.add(str(event.port))

        if event.type == "FINDING":
            location = event.data.get("url", event.data.get("host", ""))
            if location:
                self.findings.add(f"{location}:{event.data['description']}")

        if event.type == "VULNERABILITY":
            location = event.data.get("url", event.data.get("host", ""))
            if location:
                self.findings.add(f"{location}:{event.data['description']}:{event.data['severity']}")
                severity_int = severity_map.get(event.data.get("severity", "N/A"), 0)
                if severity_int > self.risk_rating:
                    self.risk_rating = severity_int

        if event.type == "TECHNOLOGY":
            self.technologies.add(event.data["technology"])

        for tag in event.tags:
            if tag.startswith("cdn-") or tag.startswith("cloud-"):
                self.provider = tag
                break

    @property
    def hostkey(self):
        return _make_hostkey(self.host, self.ip_addresses)


def _make_hostkey(host, ips):
    """
    We handle public and private IPs differently
    If the IPs are public, we dedupe by host
    If they're private, we dedupe by the IPs themselves
    """
    ips = _make_ip_list(ips)
    is_private = ips and all(is_ip(i) and i.is_private for i in ips)
    if is_private:
        return ",".join(sorted([str(i) for i in ips]))
    return str(host)


def _make_ip_list(ips):
    if isinstance(ips, str):
        ips = [i.strip() for i in ips.split(",")]
    ips = [make_ip_type(i) for i in ips if i and is_ip(i)]
    return ips



from .base import ModuleTestBase, tempwordlist


class TestMassdns(ModuleTestBase):
    subdomain_wordlist = tempwordlist(["www", "asdf"])
    config_overrides = {"modules": {"massdns": {"wordlist": str(subdomain_wordlist)}}}

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://raw.githubusercontent.com/blacklanternsecurity/public-dns-servers/master/nameservers.txt",
            text="8.8.8.8\n8.8.4.4\n1.1.1.1",
        )

    def check(self, module_test, events):
        assert any(e.data == "www.blacklanternsecurity.com" for e in events)
        assert not any(e.data == "asdf.blacklanternsecurity.com" for e in events)



from .base import ModuleTestBase


class TestHTTP(ModuleTestBase):
    downstream_url = "https://blacklanternsecurity.fakedomain:1234/events"
    config_overrides = {
        "output_modules": {
            "http": {
                "url": downstream_url,
                "method": "PUT",
                "bearer": "auth_token",
                "username": "bbot_user",
                "password": "bbot_password",
            }
        }
    }

    async def setup_after_prep(self, module_test):
        module_test.httpx_mock.add_response(
            method="PUT", headers={"Authorization": "bearer auth_token"}, url=self.downstream_url
        )

    def check(self, module_test, events):
        pass



from ..bbot_fixtures import *
from subprocess import CalledProcessError


@pytest.mark.asyncio
async def test_command(bbot_scanner, bbot_config):
    scan1 = bbot_scanner(config=bbot_config)

    # run
    assert "plumbus\n" == (await scan1.helpers.run(["echo", "plumbus"])).stdout
    assert b"plumbus\n" == (await scan1.helpers.run(["echo", "plumbus"], text=False)).stdout
    result = (await scan1.helpers.run(["cat"], input="some\nrandom\nstdin")).stdout
    assert result.splitlines() == ["some", "random", "stdin"]
    result = (await scan1.helpers.run(["cat"], input=b"some\nrandom\nstdin", text=False)).stdout
    assert result.splitlines() == [b"some", b"random", b"stdin"]
    result = (await scan1.helpers.run(["cat"], input=["some", "random", "stdin"])).stdout
    assert result.splitlines() == ["some", "random", "stdin"]
    result = (await scan1.helpers.run(["cat"], input=[b"some", b"random", b"stdin"], text=False)).stdout
    assert result.splitlines() == [b"some", b"random", b"stdin"]

    # run_live
    lines = []
    async for line in scan1.helpers.run_live(["echo", "plumbus"]):
        lines.append(line)
    assert lines == ["plumbus"]
    lines = []
    async for line in scan1.helpers.run_live(["echo", "plumbus"], text=False):
        lines.append(line)
    assert lines == [b"plumbus"]
    lines = []
    async for line in scan1.helpers.run_live(["cat"], input="some\nrandom\nstdin"):
        lines.append(line)
    assert lines == ["some", "random", "stdin"]
    lines = []
    async for line in scan1.helpers.run_live(["cat"], input=["some", "random", "stdin"]):
        lines.append(line)
    assert lines == ["some", "random", "stdin"]

    # test check=True
    with pytest.raises(CalledProcessError) as excinfo:
        lines = [l async for line in scan1.helpers.run_live(["ls", "/aslkdjflasdkfsd"], check=True)]
    assert "No such file or directory" in excinfo.value.stderr
    with pytest.raises(CalledProcessError) as excinfo:
        lines = [l async for line in scan1.helpers.run_live(["ls", "/aslkdjflasdkfsd"], check=True, text=False)]
    assert b"No such file or directory" in excinfo.value.stderr
    with pytest.raises(CalledProcessError) as excinfo:
        await scan1.helpers.run(["ls", "/aslkdjflasdkfsd"], check=True)
    assert "No such file or directory" in excinfo.value.stderr
    with pytest.raises(CalledProcessError) as excinfo:
        await scan1.helpers.run(["ls", "/aslkdjflasdkfsd"], check=True, text=False)
    assert b"No such file or directory" in excinfo.value.stderr

    # test piping
    lines = []
    async for line in scan1.helpers.run_live(
        ["cat"], input=scan1.helpers.run_live(["echo", "-en", r"some\nrandom\nstdin"])
    ):
        lines.append(line)
    assert lines == ["some", "random", "stdin"]
    lines = []
    async for line in scan1.helpers.run_live(
        ["cat"], input=scan1.helpers.run_live(["echo", "-en", r"some\nrandom\nstdin"], text=False), text=False
    ):
        lines.append(line)
    assert lines == [b"some", b"random", b"stdin"]

    # test missing executable
    result = await scan1.helpers.run(["sgkjlskdfsdf"])
    assert result is None
    lines = [l async for l in scan1.helpers.run_live(["ljhsdghsdf"])]
    assert not lines
    # test stderr
    result = await scan1.helpers.run(["ls", "/sldikgjasldkfsdf"])
    assert "No such file or directory" in result.stderr
    lines = [l async for l in scan1.helpers.run_live(["ls", "/sldikgjasldkfsdf"])]
    assert not lines

    # test sudo + existence of environment variables
    scan1.load_modules()
    path_parts = os.environ.get("PATH", "").split(":")
    assert "/tmp/.bbot_test/tools" in path_parts
    run_lines = (await scan1.helpers.run(["env"])).stdout.splitlines()
    assert "BBOT_PLUMBUS=asdf" in run_lines
    for line in run_lines:
        if line.startswith("PATH="):
            path_parts = line.split("=", 1)[-1].split(":")
            assert "/tmp/.bbot_test/tools" in path_parts
    run_lines_sudo = (await scan1.helpers.run(["env"], sudo=True)).stdout.splitlines()
    assert "BBOT_PLUMBUS=asdf" in run_lines_sudo
    for line in run_lines_sudo:
        if line.startswith("PATH="):
            path_parts = line.split("=", 1)[-1].split(":")
            assert "/tmp/.bbot_test/tools" in path_parts
    run_live_lines = [l async for l in scan1.helpers.run_live(["env"])]
    assert "BBOT_PLUMBUS=asdf" in run_live_lines
    for line in run_live_lines:
        if line.startswith("PATH="):
            path_parts = line.strip().split("=", 1)[-1].split(":")
            assert "/tmp/.bbot_test/tools" in path_parts
    run_live_lines_sudo = [l async for l in scan1.helpers.run_live(["env"], sudo=True)]
    assert "BBOT_PLUMBUS=asdf" in run_live_lines_sudo
    for line in run_live_lines_sudo:
        if line.startswith("PATH="):
            path_parts = line.strip().split("=", 1)[-1].split(":")
            assert "/tmp/.bbot_test/tools" in path_parts



import re

from ..bbot_fixtures import *


@pytest.mark.asyncio
async def test_modules_basic(scan, helpers, events, bbot_config, bbot_scanner, httpx_mock):
    fallback_nameservers = scan.helpers.temp_dir / "nameservers.txt"
    with open(fallback_nameservers, "w") as f:
        f.write("8.8.8.8\n")

    for http_method in ("GET", "CONNECT", "HEAD", "POST", "PUT", "TRACE", "DEBUG", "PATCH", "DELETE", "OPTIONS"):
        httpx_mock.add_response(method=http_method, url=re.compile(r".*"), json={"test": "test"})

    # event filtering
    from bbot.modules.base import BaseModule
    from bbot.modules.output.base import BaseOutputModule
    from bbot.modules.report.base import BaseReportModule
    from bbot.modules.internal.base import BaseInternalModule

    # output module specific event filtering tests
    base_output_module = BaseOutputModule(scan)
    base_output_module.watched_events = ["IP_ADDRESS"]
    localhost = scan.make_event("127.0.0.1", source=scan.root_event)
    assert base_output_module._event_precheck(localhost)[0] == True
    localhost._internal = True
    assert base_output_module._event_precheck(localhost)[0] == False
    localhost._force_output = True
    assert base_output_module._event_precheck(localhost)[0] == True
    localhost._omit = True
    assert base_output_module._event_precheck(localhost)[0] == False

    # common event filtering tests
    for module_class in (BaseModule, BaseOutputModule, BaseReportModule, BaseInternalModule):
        base_module = module_class(scan)
        localhost2 = scan.make_event("127.0.0.2", source=events.subdomain)
        localhost2.set_scope_distance(0)
        # base cases
        base_module._watched_events = None
        base_module.watched_events = ["*"]
        assert base_module._event_precheck(events.emoji)[0] == True
        base_module._watched_events = None
        base_module.watched_events = ["IP_ADDRESS"]
        assert base_module._event_precheck(events.ipv4)[0] == True
        assert base_module._event_precheck(events.domain)[0] == False
        assert base_module._event_precheck(events.localhost)[0] == True
        assert base_module._event_precheck(localhost2)[0] == True
        # target only
        base_module.target_only = True
        assert base_module._event_precheck(localhost2)[0] == False
        localhost2.add_tag("target")
        assert base_module._event_precheck(localhost2)[0] == True
        base_module.target_only = False
        # special case for IPs and ranges
        base_module.watched_events = ["IP_ADDRESS", "IP_RANGE"]
        ip_range = scan.make_event("127.0.0.0/24", dummy=True)
        localhost4 = scan.make_event("127.0.0.1", source=ip_range)
        localhost4.set_scope_distance(0)
        localhost4.module = "plumbus"
        assert base_module._event_precheck(localhost4)[0] == True
        localhost4.module = "speculate"
        assert base_module._event_precheck(localhost4)[0] == False

        # in scope only
        base_module.in_scope_only = True
        localhost3 = scan.make_event("127.0.0.2", source=events.subdomain)
        valid, reason = await base_module._event_postcheck(localhost3)
        if base_module._type == "output":
            assert valid
        else:
            assert not valid
            assert reason == "it did not meet in_scope_only filter criteria"
        base_module.in_scope_only = False
        base_module.scope_distance_modifier = 0
        localhost4 = scan.make_event("127.0.0.1", source=events.subdomain)
        valid, reason = await base_module._event_postcheck(events.localhost)
        assert valid

    base_output_module = BaseOutputModule(scan)
    base_output_module.watched_events = ["IP_ADDRESS"]

    scan2 = bbot_scanner(
        modules=list(set(available_modules + available_internal_modules)),
        output_modules=list(available_output_modules),
        config=bbot_config,
    )
    scan2.helpers.dns.fallback_nameservers_file = fallback_nameservers
    await scan2.load_modules()
    scan2.status = "RUNNING"

    # attributes, descriptions, etc.
    for module_name, module in sorted(scan2.modules.items()):
        # flags
        assert module._type in ("internal", "output", "scan")
        # async stuff
        not_async = []
        for func_name in ("setup", "ping", "filter_event", "handle_event", "finish", "report", "cleanup"):
            f = getattr(module, func_name)
            if not scan2.helpers.is_async_function(f):
                log.error(f"{f.__qualname__}() is not async")
                not_async.append(f)
    assert not any(not_async)

    # module preloading
    all_preloaded = module_loader.preloaded()
    assert "massdns" in all_preloaded
    assert "DNS_NAME" in all_preloaded["massdns"]["watched_events"]
    assert "DNS_NAME" in all_preloaded["massdns"]["produced_events"]
    assert "subdomain-enum" in all_preloaded["massdns"]["flags"]
    assert "wordlist" in all_preloaded["massdns"]["config"]
    assert type(all_preloaded["massdns"]["config"]["max_resolvers"]) == int
    assert all_preloaded["sslcert"]["deps"]["pip"]
    assert all_preloaded["sslcert"]["deps"]["apt"]
    assert all_preloaded["massdns"]["deps"]["ansible"]

    for module_name, preloaded in all_preloaded.items():
        # either active or passive and never both
        flags = preloaded.get("flags", [])
        if preloaded["type"] == "scan":
            assert ("active" in flags and not "passive" in flags) or (
                not "active" in flags and "passive" in flags
            ), f'module "{module_name}" must have either "active" or "passive" flag'
            assert preloaded.get("meta", {}).get("description", ""), f"{module_name} must have a description"

        # attribute checks
        watched_events = preloaded.get("watched_events")
        produced_events = preloaded.get("produced_events")

        assert type(watched_events) == list
        assert type(produced_events) == list
        if not preloaded.get("type", "") in ("internal",):
            assert watched_events, f"{module_name}.watched_events must not be empty"
        assert type(watched_events) == list, f"{module_name}.watched_events must be of type list"
        assert type(produced_events) == list, f"{module_name}.produced_events must be of type list"
        assert all(
            [type(t) == str for t in watched_events]
        ), f"{module_name}.watched_events entries must be of type string"
        assert all(
            [type(t) == str for t in produced_events]
        ), f"{module_name}.produced_events entries must be of type string"

        assert type(preloaded.get("deps_pip", [])) == list, f"{module_name}.deps_pip must be of type list"
        assert (
            type(preloaded.get("deps_pip_constraints", [])) == list
        ), f"{module_name}.deps_pip_constraints must be of type list"
        assert type(preloaded.get("deps_apt", [])) == list, f"{module_name}.deps_apt must be of type list"
        assert type(preloaded.get("deps_shell", [])) == list, f"{module_name}.deps_shell must be of type list"
        assert type(preloaded.get("config", None)) == dict, f"{module_name}.options must be of type list"
        assert type(preloaded.get("options_desc", None)) == dict, f"{module_name}.options_desc must be of type list"
        # options must have descriptions
        assert set(preloaded.get("config", {})) == set(
            preloaded.get("options_desc", {})
        ), f"{module_name}.options do not match options_desc"
        # descriptions most not be blank
        assert all(
            o for o in preloaded.get("options_desc", {}).values()
        ), f"{module_name}.options_desc descriptions must not be blank"



from bbot.modules.base import BaseModule
from wafw00f import main as wafw00f_main


class wafw00f(BaseModule):
    """
    https://github.com/EnableSecurity/wafw00f
    """

    watched_events = ["URL"]
    produced_events = ["WAF"]
    flags = ["active", "aggressive"]
    meta = {"description": "Web Application Firewall Fingerprinting Tool"}

    deps_pip = ["wafw00f~=2.2.0"]

    options = {"generic_detect": True}
    options_desc = {"generic_detect": "When no specific WAF detections are made, try to peform a generic detect"}

    in_scope_only = True

    async def setup(self):
        self.scanned_hosts = set()
        return True

    async def handle_event(self, event):
        parsed_host = event.parsed
        host = f"{parsed_host.scheme}://{parsed_host.netloc}/"
        host_hash = hash(host)
        if host_hash in self.scanned_hosts:
            self.debug(f"Host {host} was already scanned, exiting")
            return
        else:
            self.scanned_hosts.add(host_hash)

        WW = await self.scan.run_in_executor(wafw00f_main.WAFW00F, host)
        waf_detections = await self.scan.run_in_executor(WW.identwaf)
        if waf_detections:
            for waf in waf_detections:
                self.emit_event({"host": host, "WAF": waf}, "WAF", source=event)
        else:
            if self.config.get("generic_detect") == True:
                generic = await self.scan.run_in_executor(WW.genericdetect)
                if generic:
                    self.emit_event(
                        {"host": host, "WAF": "generic detection", "info": WW.knowledge["generic"]["reason"]},
                        "WAF",
                        source=event,
                    )



from .test_module_bucket_aws import *


class TestBucket_Azure(Bucket_AWS_Base):
    provider = "azure"
    random_bucket_1 = f"{random_bucket_name_1}.blob.core.windows.net"
    random_bucket_2 = f"{random_bucket_name_2}.blob.core.windows.net"
    random_bucket_3 = f"{random_bucket_name_3}.blob.core.windows.net"

    def url_setup(self):
        self.url_1 = f"https://{self.random_bucket_1}"
        self.url_2 = f"https://{self.random_bucket_2}"
        self.url_3 = f"https://{self.random_bucket_3}/{random_bucket_name_3}?restype=container"
